"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getChallengeClearedEvent = exports.getChallengeRegisteredEvent = exports.hashChallengeMessage = void 0;
const ethers_1 = require("ethers");
const { Interface, keccak256, defaultAbiCoder } = ethers_1.utils;
const NitroAdjudicator_json_1 = __importDefault(require("../../artifacts/contracts/NitroAdjudicator.sol/NitroAdjudicator.json"));
const outcome_1 = require("./outcome");
const state_1 = require("./state");
function hashChallengeMessage(challengeState) {
    return keccak256(defaultAbiCoder.encode(['bytes32', 'string'], [(0, state_1.hashState)(challengeState), 'forceMove']));
}
exports.hashChallengeMessage = hashChallengeMessage;
/**
 * Extracts a ChallengeRegisteredEvent (containing challengeStates) from the supplied eventResult.
 * @param eventResult the event itself
 * @returns a ChallengeRegisteredEvent
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function getChallengeRegisteredEvent(eventResult) {
    const { channelId, turnNumRecord, finalizesAt, isFinal, fixedPart, variableParts: variablePartsUnstructured, sigs, } = eventResult.slice(-1)[0].args;
    // Fixed part
    const participants = fixedPart[1].map((p) => ethers_1.BigNumber.from(p).toHexString());
    const channelNonce = fixedPart[2];
    const appDefinition = fixedPart[3];
    const challengeDuration = ethers_1.BigNumber.from(fixedPart[4]).toNumber();
    // Variable part
    const variableParts = variablePartsUnstructured.map(v => {
        const [outcome, appData, turnNum, isFinal] = v;
        return { outcome, appData, turnNum, isFinal };
    });
    const challengeStates = variableParts.map((v, i) => {
        const turnNum = turnNumRecord - (variableParts.length - i - 1);
        const signature = sigs[i];
        const state = {
            turnNum,
            channelNonce,
            participants,
            outcome: (0, outcome_1.decodeOutcome)(v.outcome),
            appData: v.appData,
            challengeDuration,
            appDefinition,
            isFinal,
        };
        return { state, signature };
    });
    return { channelId, challengeStates, finalizesAt };
}
exports.getChallengeRegisteredEvent = getChallengeRegisteredEvent;
/**
 * Extracts a ChallengeClearedEvent (containing a new signedState) from the logs of a respond or checkpoint transaction
 * @param tx a suitable transaction causing a ChallengeCleared event to be emitted
 * @param eventResult the event itself
 * @returns a ChallengeClearedEvent
 */
function getChallengeClearedEvent(tx, 
// eslint-disable-next-line @typescript-eslint/no-explicit-any
eventResult) {
    const { newTurnNumRecord } = eventResult.slice(-1)[0].args;
    // https://github.com/ethers-io/ethers.js/issues/602#issuecomment-574671078
    const decodedTransaction = new Interface(NitroAdjudicator_json_1.default.abi).parseTransaction(tx);
    if (decodedTransaction.name === 'respond') {
        // NOTE: args value is an array of the inputted arguments, not an object with labelled keys
        // ethers.js should change this, and when it does, we can use the commented out type
        const args /* RespondTransactionArguments */ = decodedTransaction.args;
        const [participants, channelNonce, appDefinition, challengeDuration] = args[2];
        const isFinal = args[1][1];
        const outcome = (0, outcome_1.decodeOutcome)(args[3][1][0]);
        const appData = args[3][1][1];
        const signature = {
            v: args[4][0],
            r: args[4][1],
            s: args[4][2],
            _vs: args[4][3],
            recoveryParam: args[4][4],
        };
        const signedState = {
            signature,
            state: {
                challengeDuration,
                appDefinition,
                isFinal,
                outcome,
                appData,
                channelNonce,
                participants,
                turnNum: ethers_1.BigNumber.from(newTurnNumRecord).toNumber(),
            },
        };
        return {
            kind: 'respond',
            newStates: [signedState],
        };
    }
    else if (decodedTransaction.name === 'checkpoint') {
        throw new Error('UnimplementedError');
    }
    else {
        throw new Error('Unexpected call to getChallengeClearedEvent with invalid or unrelated transaction data');
    }
}
exports.getChallengeClearedEvent = getChallengeClearedEvent;
