"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.encodeFingerprintPreimage = exports.parseStatus = exports.channelDataToStatus = void 0;
const ethers_1 = require("ethers");
const outcome_1 = require("./outcome");
const state_1 = require("./state");
const FINGERPRINT_PREIMAGE_TYPE = `tuple(
  bytes32 stateHash,
  bytes32 outcomeHash
)`;
/**
 * Computes the on chain status from the supplied channelData
 * @param channelData
 * @returns the 32 byte "status" word that may be stored on chain for this channel
 */
function channelDataToStatus(channelData) {
    const { turnNumRecord, finalizesAt } = channelData;
    const hash = ethers_1.utils.keccak256(encodeFingerprintPreimage(channelData));
    const fingerprint = ethers_1.utils.hexDataSlice(hash, 12);
    const status = '0x' +
        ethers_1.utils.hexZeroPad(ethers_1.utils.hexlify(turnNumRecord), 6).slice(2) +
        ethers_1.utils.hexZeroPad(ethers_1.utils.hexlify(finalizesAt), 6).slice(2) +
        fingerprint.slice(2);
    return status;
}
exports.channelDataToStatus = channelDataToStatus;
function parseStatus(status) {
    validateHexString(status);
    //
    let cursor = 2;
    const turnNumRecord = '0x' + status.slice(cursor, (cursor += 12));
    const finalizesAt = '0x' + status.slice(cursor, (cursor += 12));
    const fingerprint = '0x' + status.slice(cursor);
    return {
        turnNumRecord: asNumber(turnNumRecord),
        finalizesAt: asNumber(finalizesAt),
        fingerprint,
    };
}
exports.parseStatus = parseStatus;
const asNumber = s => ethers_1.BigNumber.from(s).toNumber();
function getFingerprintPreimage({ finalizesAt, state, outcome }) {
    /*
    When the channel is not open, it is still possible for the state to be missing, indicating that the channel is finalized.
    It is currently up to the caller to ensure this.
    */
    const isOpen = finalizesAt === 0;
    if (isOpen && (outcome || state)) {
        console.warn(`Invalid open channel storage: ${JSON.stringify(outcome || state)}`);
    }
    const stateHash = isOpen || !state ? ethers_1.constants.HashZero : (0, state_1.hashState)(state);
    const outcomeHash = isOpen || !outcome ? ethers_1.constants.HashZero : (0, outcome_1.hashOutcome)(outcome);
    return { stateHash, outcomeHash };
}
function encodeFingerprintPreimage(data) {
    return ethers_1.utils.defaultAbiCoder.encode([FINGERPRINT_PREIMAGE_TYPE], [getFingerprintPreimage(data)]);
}
exports.encodeFingerprintPreimage = encodeFingerprintPreimage;
function validateHexString(hexString) {
    if (!ethers_1.utils.isHexString(hexString)) {
        throw new Error(`Not a hex string: ${hexString}`);
    }
    if (hexString.length !== 66) {
        throw new Error(`Incorrect length: ${hexString.length}`);
    }
}
