"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.computeTransferEffectsAndInteractions = exports.computeReclaimEffects = exports.convertAddressToBytes32 = exports.convertBytes32ToAddress = exports.getDepositedEvent = void 0;
const ethers_1 = require("ethers");
const exit_format_1 = require("@statechannels/exit-format");
const ethers_utils_1 = require("../ethers-utils");
const outcome_1 = require("./outcome");
/**
 * Extracts a DepositedEvent from a suitable eventResult
 * @param eventResult
 * @returns a DepositedEvent
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function getDepositedEvent(eventResult) {
    const { destination, destinationHoldings } = (0, ethers_utils_1.parseEventResult)(eventResult);
    return {
        destination,
        destinationHoldings: ethers_1.BigNumber.from(destinationHoldings),
    };
}
exports.getDepositedEvent = getDepositedEvent;
/**
 * Extracts a 20 byte address hex string from a 32 byte hex string by slicing
 * @param bytes32 a hex string
 * @returns a copy of the hex string with the first 12 bytes removed (after the "0x")
 */
function convertBytes32ToAddress(bytes32) {
    const normalized = ethers_1.utils.hexZeroPad(bytes32, 32);
    return ethers_1.utils.getAddress(`0x${normalized.slice(-40)}`);
}
exports.convertBytes32ToAddress = convertBytes32ToAddress;
/**
 * Left pads a 20 byte address hex string with zeros until it is a 32 byte hex string
 * e.g.,
 * 0x9546E319878D2ca7a21b481F873681DF344E0Df8 becomes
 * 0x0000000000000000000000009546E319878D2ca7a21b481F873681DF344E0Df8
 * @param address 20 byte hex string
 * @returns 32 byte padded hex string
 */
function convertAddressToBytes32(address) {
    const normalizedAddress = ethers_1.BigNumber.from(address).toHexString();
    if (!ethers_1.utils.isAddress(address)) {
        throw new Error(`Input is not a valid Ethereum address.`);
    }
    // We pad to 66 = (32*2) + 2('0x')
    return ethers_1.utils.hexZeroPad(normalizedAddress, 32);
}
exports.convertAddressToBytes32 = convertAddressToBytes32;
/**
 *
 * Emulates solidity code. TODO replace with PureEVM implementation?
 * @param initialHoldings
 * @param allocation
 * @param indices
 */
function computeReclaimEffects(sourceAllocations, // we must index this with a JS number that is less than 2**32 - 1
targetAllocations, // we must index this with a JS number that is less than 2**32 - 1
indexOfTargetInSource) {
    const newSourceAllocations = []; // will be one slot shorter than sourceAllocations
    const guarantee = sourceAllocations[indexOfTargetInSource];
    if (guarantee.allocationType != exit_format_1.AllocationType.guarantee) {
        throw Error('not a guarantee');
    }
    const { left, right } = (0, outcome_1.decodeGuaranteeData)(guarantee.metadata);
    let foundTarget = false;
    let foundLeft = false;
    let foundRight = false;
    let totalReclaimed = ethers_1.BigNumber.from(0);
    let k = 0;
    for (let i = 0; i < sourceAllocations.length; i++) {
        if (i == indexOfTargetInSource) {
            foundTarget = true;
            continue;
        }
        newSourceAllocations[k] = {
            destination: sourceAllocations[i].destination,
            amount: sourceAllocations[i].amount,
            allocationType: sourceAllocations[i].allocationType,
            metadata: sourceAllocations[i].metadata,
        };
        // copy each element except the indexOfTargetInSource element
        if (!foundLeft && sourceAllocations[i].destination.toLowerCase() == left.toLowerCase()) {
            newSourceAllocations[k].amount = ethers_1.BigNumber.from(sourceAllocations[i].amount)
                .add(targetAllocations[0].amount)
                .toHexString();
            totalReclaimed = totalReclaimed.add(targetAllocations[0].amount);
            foundLeft = true;
        }
        if (!foundRight && sourceAllocations[i].destination.toLowerCase() == right.toLowerCase()) {
            newSourceAllocations[k].amount = ethers_1.BigNumber.from(sourceAllocations[i].amount)
                .add(targetAllocations[1].amount)
                .toHexString();
            totalReclaimed = totalReclaimed.add(targetAllocations[1].amount);
            foundRight = true;
        }
        k++;
    }
    if (!foundTarget) {
        throw Error('could not find target');
    }
    if (!foundLeft) {
        throw Error('could not find left');
    }
    if (!foundRight) {
        throw Error('could not find right');
    }
    if (!totalReclaimed.eq(guarantee.amount)) {
        throw Error('totalReclaimed!=guarantee.amount');
    }
    return newSourceAllocations;
}
exports.computeReclaimEffects = computeReclaimEffects;
/**
 *
 * Emulates solidity code. TODO replace with PureEVM implementation?
 * @param initialHoldings
 * @param allocation
 * @param indices
 */
function computeTransferEffectsAndInteractions(initialHoldings, allocations, // we must index this with a JS number that is less than 2**32 - 1
indices) {
    let totalPayouts = ethers_1.BigNumber.from(0);
    const newAllocations = [];
    const exitAllocations = Array(indices.length > 0 ? indices.length : allocations.length).fill({
        destination: ethers_1.constants.HashZero,
        amount: '0x00',
        metadata: '0x',
        allocationType: 0,
    });
    let allocatesOnlyZeros = true;
    let surplus = ethers_1.BigNumber.from(initialHoldings);
    let k = 0;
    for (let i = 0; i < allocations.length; i++) {
        newAllocations.push({
            destination: allocations[i].destination,
            amount: ethers_1.BigNumber.from(0).toHexString(),
            metadata: allocations[i].metadata,
            allocationType: allocations[i].allocationType,
        });
        const affordsForDestination = min(ethers_1.BigNumber.from(allocations[i].amount), surplus);
        if (indices.length == 0 || (k < indices.length && indices[k] === i)) {
            newAllocations[i].amount = ethers_1.BigNumber.from(allocations[i].amount)
                .sub(affordsForDestination)
                .toHexString();
            exitAllocations[k] = {
                destination: allocations[i].destination,
                amount: affordsForDestination.toHexString(),
                metadata: allocations[i].metadata,
                allocationType: allocations[i].allocationType,
            };
            totalPayouts = totalPayouts.add(affordsForDestination);
            ++k;
        }
        else {
            newAllocations[i].amount = allocations[i].amount;
        }
        if (!ethers_1.BigNumber.from(newAllocations[i].amount).isZero())
            allocatesOnlyZeros = false;
        surplus = surplus.sub(affordsForDestination);
    }
    return {
        newAllocations,
        allocatesOnlyZeros,
        exitAllocations,
        totalPayouts: totalPayouts.toHexString(),
    };
}
exports.computeTransferEffectsAndInteractions = computeTransferEffectsAndInteractions;
function min(a, b) {
    return a.gt(b) ? b : a;
}
