import { Signature } from '@ethersproject/bytes';
import { Outcome } from './outcome';
import { Address, Bytes, Bytes32, Uint256, Uint48, Uint64 } from './types';
export type State = FixedPart & VariablePart;
/**
 * The part of a State which does not ordinarily change during state channel updates
 */
export interface FixedPart {
    participants: Address[];
    channelNonce: Uint64;
    appDefinition: Address;
    challengeDuration: Uint48;
}
/**
 * The part of a State which usually changes during state channel updates
 */
export interface VariablePart {
    outcome: Outcome;
    appData: Bytes;
    turnNum: Uint48;
    isFinal: boolean;
}
/**
 * Variable part with its signatures created by participants
 */
export interface SignedVariablePart {
    variablePart: VariablePart;
    sigs: Signature[];
}
/**
 * Variable part with a signedBy bitfield.
 */
export interface RecoveredVariablePart {
    variablePart: VariablePart;
    signedBy: Uint256;
}
/**
 * Separates supplied array into proof and candidate.
 * @param svps Array of SignedVariableParts or RecoveredVariableParts.
 * @returns proof and candidate, of the same type as input.
 */
export declare function separateProofAndCandidate<T = SignedVariablePart | RecoveredVariablePart>(svps: T[]): {
    proof: T[];
    candidate: T;
};
/**
 * Extracts the VariablePart of a state
 * @param state a State
 * @returns the VariablePart, which usually changes during state channel updates
 */
export declare function getVariablePart(state: State): VariablePart;
/**
 * Extracts the FixedPart of a state
 * @param state a State
 * @returns the FixedPart, which does not ordinarily change during state channel updates
 */
export declare function getFixedPart(state: State): FixedPart;
/**
 * Encodes and hashes the AppPart of a state
 * @param state a State
 * @returns a 32 byte keccak256 hash
 */
export declare function hashAppPart(state: State): Bytes32;
/**
 * Encodes a state
 * @param state a State
 * @returns bytes array encoding
 */
export declare function encodeState(state: State): Bytes;
/**
 * Hashes a state
 * @param state a State
 * @returns a 32 byte keccak256 hash
 */
export declare function hashState(state: State): Bytes32;
