"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.hashState = exports.encodeState = exports.hashAppPart = exports.getFixedPart = exports.getVariablePart = exports.separateProofAndCandidate = void 0;
const ethers_1 = require("ethers");
const channel_1 = require("./channel");
/**
 * Separates supplied array into proof and candidate.
 * @param svps Array of SignedVariableParts or RecoveredVariableParts.
 * @returns proof and candidate, of the same type as input.
 */
function separateProofAndCandidate(svps) {
    const proof = svps.slice(0, -1);
    const candidate = svps[svps.length - 1];
    if (candidate == undefined) {
        throw Error('insufficient length array');
    }
    return { proof, candidate };
}
exports.separateProofAndCandidate = separateProofAndCandidate;
/**
 * Extracts the VariablePart of a state
 * @param state a State
 * @returns the VariablePart, which usually changes during state channel updates
 */
function getVariablePart(state) {
    return {
        outcome: state.outcome,
        appData: state.appData,
        turnNum: state.turnNum,
        isFinal: state.isFinal,
    };
}
exports.getVariablePart = getVariablePart;
/**
 * Extracts the FixedPart of a state
 * @param state a State
 * @returns the FixedPart, which does not ordinarily change during state channel updates
 */
function getFixedPart(state) {
    const { appDefinition, challengeDuration, participants, channelNonce } = state;
    return { participants, channelNonce, appDefinition, challengeDuration };
}
exports.getFixedPart = getFixedPart;
/**
 * Encodes and hashes the AppPart of a state
 * @param state a State
 * @returns a 32 byte keccak256 hash
 */
function hashAppPart(state) {
    const { challengeDuration, appDefinition, appData } = state;
    return ethers_1.utils.keccak256(ethers_1.utils.defaultAbiCoder.encode(['uint256', 'address', 'bytes'], [challengeDuration, appDefinition, appData]));
}
exports.hashAppPart = hashAppPart;
/**
 * Encodes a state
 * @param state a State
 * @returns bytes array encoding
 */
function encodeState(state) {
    const { turnNum, isFinal, appData, outcome } = state;
    const channelId = (0, channel_1.getChannelId)(getFixedPart(state));
    return ethers_1.utils.defaultAbiCoder.encode([
        'bytes32',
        'bytes',
        {
            type: 'tuple[]',
            components: [
                { name: 'asset', type: 'address' },
                {
                    name: 'assetMetadata',
                    type: 'tuple',
                    components: [
                        { name: 'assetType', type: 'uint8' },
                        { name: 'metadata', type: 'bytes' },
                    ],
                },
                {
                    type: 'tuple[]',
                    name: 'allocations',
                    components: [
                        { name: 'destination', type: 'bytes32' },
                        { name: 'amount', type: 'uint256' },
                        { name: 'allocationType', type: 'uint8' },
                        { name: 'metadata', type: 'bytes' },
                    ],
                },
            ],
        },
        'uint256',
        'bool',
    ], [channelId, appData, outcome, turnNum, isFinal]);
}
exports.encodeState = encodeState;
/**
 * Hashes a state
 * @param state a State
 * @returns a 32 byte keccak256 hash
 */
function hashState(state) {
    return ethers_1.utils.keccak256(encodeState(state));
}
exports.hashState = hashState;
