"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.concludeArgs = exports.createConcludeTransaction = exports.checkpointArgs = exports.createCheckpointTransaction = exports.createChallengeTransaction = void 0;
const signatures_1 = require("../../signatures");
const state_1 = require("../state");
const multi_asset_holder_1 = require("./multi-asset-holder");
function createChallengeTransaction(states, // in turnNum order [..,state-with-largestTurnNum]
signatures, // in participant order: [sig-from-p0, sig-from-p1, ...]
whoSignedWhat, challengerPrivateKey) {
    // Sanity checks on expected lengths
    if (states.length === 0) {
        throw new Error('No states provided');
    }
    const { participants } = states[0];
    if (participants.length !== signatures.length) {
        throw new Error(`Participants (length:${participants.length}) and signatures (length:${signatures.length}) need to be the same length`);
    }
    const fixedPart = (0, state_1.getFixedPart)(states[0]);
    const variableParts = states.map(s => (0, state_1.getVariablePart)(s));
    const { proof, candidate } = (0, state_1.separateProofAndCandidate)((0, signatures_1.bindSignatures)(variableParts, signatures, whoSignedWhat));
    // Q: Is there a reason why createForceMoveTransaction accepts a State[] and a Signature[]
    // Argument rather than a SignedState[] argument?
    // A: Yes, because the signatures must be passed in participant order: [sig-from-p0, sig-from-p1, ...]
    // and SignedStates[] won't comply with that in general. This function accepts the re-ordered sigs.
    const signedStates = states.map(s => ({
        state: s,
        signature: { v: 0, r: '', s: '', _vs: '', recoveryParam: 0 },
    }));
    const challengerSignature = (0, signatures_1.signChallengeMessage)(signedStates, challengerPrivateKey);
    const data = multi_asset_holder_1.NitroAdjudicatorContractInterface.encodeFunctionData('challenge', [
        fixedPart,
        proof,
        candidate,
        challengerSignature,
    ]);
    return { data };
}
exports.createChallengeTransaction = createChallengeTransaction;
function createCheckpointTransaction({ states, signatures, whoSignedWhat, }) {
    const data = multi_asset_holder_1.NitroAdjudicatorContractInterface.encodeFunctionData('checkpoint', checkpointArgs({ states, signatures, whoSignedWhat }));
    return { data };
}
exports.createCheckpointTransaction = createCheckpointTransaction;
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function checkpointArgs({ states, signatures, whoSignedWhat }) {
    const fixedPart = (0, state_1.getFixedPart)(states[0]);
    const variableParts = states.map(s => (0, state_1.getVariablePart)(s));
    const { proof, candidate } = (0, state_1.separateProofAndCandidate)((0, signatures_1.bindSignatures)(variableParts, signatures, whoSignedWhat));
    return [fixedPart, proof, candidate];
}
exports.checkpointArgs = checkpointArgs;
function createConcludeTransaction(states, signatures, whoSignedWhat) {
    const data = multi_asset_holder_1.NitroAdjudicatorContractInterface.encodeFunctionData('conclude', concludeArgs(states, signatures, whoSignedWhat));
    return { data };
}
exports.createConcludeTransaction = createConcludeTransaction;
function concludeArgs(states, signatures, whoSignedWhat
// eslint-disable-next-line @typescript-eslint/no-explicit-any
) {
    // Sanity checks on expected lengths
    if (states.length === 0) {
        throw new Error('No states provided');
    }
    const { participants } = states[0];
    if (participants.length !== signatures.length) {
        throw new Error(`Participants (length:${participants.length}) and signatures (length:${signatures.length}) need to be the same length`);
    }
    const fixedPart = (0, state_1.getFixedPart)(states[0]);
    const variableParts = states.map(s => (0, state_1.getVariablePart)(s));
    const { candidate } = (0, state_1.separateProofAndCandidate)((0, signatures_1.bindSignatures)(variableParts, signatures, whoSignedWhat));
    return [fixedPart, candidate];
}
exports.concludeArgs = concludeArgs;
