import { Wallet, Signature } from 'ethers';
import { Bytes, Uint256 } from './contract/types';
import { Outcome } from './contract/outcome';
import { RecoveredVariablePart, SignedVariablePart, State, VariablePart } from './contract/state';
/**
 * A {@link State} along with a {@link Signature} on it
 */
export interface SignedState {
    state: State;
    signature: Signature;
}
export declare function getStateSignerAddress(signedState: SignedState): string;
/**
 * Encodes, hashes and signs a State using the supplied privateKey
 * @param state a State
 * @param privateKey an ECDSA private key
 * @returns a SignedState
 */
export declare function signState(state: State, privateKey: string): SignedState;
export declare function sign(wallet: Wallet, msgHash: string | Uint8Array): Promise<Signature>;
export type ShortenedVariablePart = {
    signerIndices: number[];
    appData?: Bytes;
    outcome?: Outcome;
    isFinal?: boolean;
} | number[];
export type TurnNumToShortenedVariablePart = Map<number, ShortenedVariablePart>;
/**
 * Signs a state using wallets at indices determined by signedBy bitfield.
 * @param state A state to sign.
 * @param wallets An array of wallets.
 * @param signedBy SignedBy bitfield.
 * @returns States signed by wallets at indices determined by signedBy bitfield.
 */
export declare function signStateWithBitfield(state: State, wallets: Wallet[], signedBy: Uint256): Promise<Signature[]>;
/**
 * Convert provided ShortenedSignedBy and turnNum into RecoveredVariablePart. Missing ShortenedVariablePart fields are initialized to default values.
 * @param turnNum Turn number.
 * @param shortenedVP ShortenedVariablePart.
 * @returns RecoveredVariablePart.
 */
export declare function shortenedToRecoveredVariablePart(turnNum: number, shortenedVP: ShortenedVariablePart): RecoveredVariablePart;
/**
 * Convert a mapping between turnNumbers and ShortenedVariableParts to an array of RecoveredVariableParts.
 * @param turnNumToShortenedVP A mapping between turnNumbers and ShortenedVariableParts.
 * @returns An array of RecoveredVariableParts.
 */
export declare function shortenedToRecoveredVariableParts(turnNumToShortenedVP: TurnNumToShortenedVariablePart): RecoveredVariablePart[];
/**
 * Maps the supplied wallets array to (a Promise of) an array of signatures by those wallets on the supplied states, using whoSignedWhat to map from wallet to state.
 */
export declare function signStates(states: State[], wallets: Wallet[], whoSignedWhat: number[]): Promise<Signature[]>;
/**
 * Maps supplied signatures to variable parts, using whoSignedWhat.
 */
export declare function bindSignatures(variableParts: VariablePart[], signatures: Signature[], whoSignedWhat: number[]): SignedVariablePart[];
/**
 * Maps supplied signatures to variable parts, using whoSignedWhat.
 */
export declare function bindSignaturesWithSignedByBitfield(variableParts: VariablePart[], signatures: Signature[], whoSignedWhat: number[]): RecoveredVariablePart[];
/**
 * Signs a challenge message (necessary for submitting a challenge) using the last of the supplied signedStates and privateKey
 * @param signedStates an array of type SignedState
 * @param privateKey an ECDSA private key (must be a participant in the channel)
 * @returns a Signature
 */
export declare function signChallengeMessage(signedStates: SignedState[], privateKey: string): Signature;
export declare function signData(hashedData: string, privateKey: string): Signature;
