"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.signData = exports.signChallengeMessage = exports.bindSignaturesWithSignedByBitfield = exports.bindSignatures = exports.signStates = exports.shortenedToRecoveredVariableParts = exports.shortenedToRecoveredVariablePart = exports.signStateWithBitfield = exports.sign = exports.signState = exports.getStateSignerAddress = void 0;
const ethers_1 = require("ethers");
const bitfield_utils_1 = require("./bitfield-utils");
const challenge_1 = require("./contract/challenge");
const channel_1 = require("./contract/channel");
const state_1 = require("./contract/state");
function getStateSignerAddress(signedState) {
    const stateHash = (0, state_1.hashState)(signedState.state);
    const recoveredAddress = ethers_1.utils.verifyMessage(ethers_1.utils.arrayify(stateHash), signedState.signature);
    const participants = signedState.state.participants;
    if (participants.indexOf(recoveredAddress) < 0) {
        throw new Error(`Recovered address ${recoveredAddress} is not a participant in channel ${(0, channel_1.getChannelId)((0, state_1.getFixedPart)(signedState.state))}`);
    }
    return recoveredAddress;
}
exports.getStateSignerAddress = getStateSignerAddress;
/**
 * Encodes, hashes and signs a State using the supplied privateKey
 * @param state a State
 * @param privateKey an ECDSA private key
 * @returns a SignedState
 */
function signState(state, privateKey) {
    const wallet = new ethers_1.Wallet(privateKey);
    if (state.participants.indexOf(wallet.address) < 0) {
        throw new Error("The state must be signed with a participant's private key");
    }
    const hashedState = (0, state_1.hashState)(state);
    const signature = signData(hashedState, privateKey);
    return { state, signature };
}
exports.signState = signState;
async function sign(wallet, msgHash) {
    // MsgHash is a hex string
    // Returns an object with v, r, and s properties.
    return ethers_1.utils.splitSignature(await wallet.signMessage(ethers_1.utils.arrayify(msgHash)));
}
exports.sign = sign;
/**
 * Signs a state using wallets at indices determined by signedBy bitfield.
 * @param state A state to sign.
 * @param wallets An array of wallets.
 * @param signedBy SignedBy bitfield.
 * @returns States signed by wallets at indices determined by signedBy bitfield.
 */
async function signStateWithBitfield(state, wallets, signedBy) {
    if (wallets.length < (0, bitfield_utils_1.getSignersNum)(signedBy)) {
        throw new Error('Insufficient wallets');
    }
    const promises = (0, bitfield_utils_1.getSignersIndices)(signedBy).map(async (i) => await sign(wallets[i], (0, state_1.hashState)(state)));
    return Promise.all(promises);
}
exports.signStateWithBitfield = signStateWithBitfield;
/**
 * Convert provided ShortenedSignedBy and turnNum into RecoveredVariablePart. Missing ShortenedVariablePart fields are initialized to default values.
 * @param turnNum Turn number.
 * @param shortenedVP ShortenedVariablePart.
 * @returns RecoveredVariablePart.
 */
function shortenedToRecoveredVariablePart(turnNum, shortenedVP) {
    var _a, _b, _c;
    // if just an array of signerIndices was supplied, convert it to ShortenedVariablePart
    if (Array.isArray(shortenedVP)) {
        shortenedVP = {
            signerIndices: shortenedVP,
        };
    }
    const outcome = (_a = shortenedVP.outcome) !== null && _a !== void 0 ? _a : [];
    const appData = (_b = shortenedVP.appData) !== null && _b !== void 0 ? _b : '0x';
    const isFinal = (_c = shortenedVP.isFinal) !== null && _c !== void 0 ? _c : false;
    return {
        variablePart: {
            outcome,
            appData,
            turnNum,
            isFinal,
        },
        signedBy: (0, bitfield_utils_1.getSignedBy)(shortenedVP.signerIndices),
    };
}
exports.shortenedToRecoveredVariablePart = shortenedToRecoveredVariablePart;
/**
 * Convert a mapping between turnNumbers and ShortenedVariableParts to an array of RecoveredVariableParts.
 * @param turnNumToShortenedVP A mapping between turnNumbers and ShortenedVariableParts.
 * @returns An array of RecoveredVariableParts.
 */
function shortenedToRecoveredVariableParts(turnNumToShortenedVP) {
    return Array.from(turnNumToShortenedVP).map(([turnNum, shortenedVP]) => {
        return shortenedToRecoveredVariablePart(turnNum, shortenedVP);
    });
}
exports.shortenedToRecoveredVariableParts = shortenedToRecoveredVariableParts;
//
// end towards #761 https://github.com/statechannels/go-nitro/issues/761
/**
 * Maps the supplied wallets array to (a Promise of) an array of signatures by those wallets on the supplied states, using whoSignedWhat to map from wallet to state.
 */
async function signStates(states, wallets, whoSignedWhat) {
    const stateHashes = states.map(s => (0, state_1.hashState)(s));
    const promises = wallets.map(async (w, i) => await sign(w, stateHashes[whoSignedWhat[i]]));
    return Promise.all(promises);
}
exports.signStates = signStates;
/**
 * Maps supplied signatures to variable parts, using whoSignedWhat.
 */
function bindSignatures(variableParts, signatures, whoSignedWhat) {
    const signedVariableParts = variableParts.map(vp => ({
        variablePart: vp,
        sigs: [],
    }));
    for (let i = 0; i < signatures.length; i++) {
        signedVariableParts[whoSignedWhat[i]].sigs.push(signatures[i]);
    }
    return signedVariableParts;
}
exports.bindSignatures = bindSignatures;
/**
 * Maps supplied signatures to variable parts, using whoSignedWhat.
 */
function bindSignaturesWithSignedByBitfield(variableParts, signatures, whoSignedWhat) {
    const recoveredVariableParts = variableParts.map(vp => ({
        variablePart: vp,
        signedBy: '0',
    }));
    for (let i = 0; i < signatures.length; i++) {
        const updatedSignedBy = Number(recoveredVariableParts[whoSignedWhat[i]].signedBy) | (2 ** i);
        recoveredVariableParts[whoSignedWhat[i]].signedBy = updatedSignedBy.toString();
    }
    return recoveredVariableParts;
}
exports.bindSignaturesWithSignedByBitfield = bindSignaturesWithSignedByBitfield;
/**
 * Signs a challenge message (necessary for submitting a challenge) using the last of the supplied signedStates and privateKey
 * @param signedStates an array of type SignedState
 * @param privateKey an ECDSA private key (must be a participant in the channel)
 * @returns a Signature
 */
function signChallengeMessage(signedStates, privateKey) {
    if (signedStates.length === 0) {
        throw new Error('At least one signed state must be provided');
    }
    const wallet = new ethers_1.Wallet(privateKey);
    if (signedStates[0].state.participants.indexOf(wallet.address) < 0) {
        throw new Error("The state must be signed with a participant's private key");
    }
    const challengeState = signedStates[signedStates.length - 1].state;
    const challengeHash = (0, challenge_1.hashChallengeMessage)(challengeState);
    return signData(challengeHash, privateKey);
}
exports.signChallengeMessage = signChallengeMessage;
function hashMessage(hashedData) {
    return ethers_1.utils.hashMessage(ethers_1.utils.arrayify(hashedData));
}
function signData(hashedData, privateKey) {
    const signingKey = new ethers_1.utils.SigningKey(privateKey);
    return ethers_1.utils.splitSignature(signingKey.signDigest(hashMessage(hashedData)));
}
exports.signData = signData;
