import { providers, Signature } from 'ethers';
import { State } from './contract/state';
import { SignedState } from './signatures';
export declare const MAGIC_ADDRESS_INDICATING_ETH = "0x0000000000000000000000000000000000000000";
export declare const MAX_TX_DATA_SIZE: number;
export declare const NITRO_MAX_GAS = 6000000;
/**
 * Marshalls the supplied signedStates into an ethereum transaction for the challenge method. Automatically computes signedVariableParts, etc.
 * @param signedStates an array of signed states
 * @param challengePrivateKey private key of the challenger
 * @returns An ethers TransactionRequest. This can be launched with `await signer.sendTransaction({to: adjudicator.address, ...txRequest}`)
 */
export declare function createChallengeTransaction(signedStates: SignedState[], challengePrivateKey: string): providers.TransactionRequest;
/**
 * Marshalls the supplied signedStates into an ethereum transaction for the checkpoint method. Automatically computes signedVariableParts, etc.
 * @param signedStates an array of signed states
 * @returns An ethers TransactionRequest. This can be launched with `await signer.sendTransaction({to: adjudicator.address, ...txRequest}`)
 */
export declare function createCheckpointTransaction(signedStates: SignedState[]): providers.TransactionRequest;
/**
 * Marshalls the supplied signedStates into an ethereum transaction for the conclude method. Automatically computes signedVariableParts, etc.
 * @param conclusionProof an array of signed states
 * @returns An ethers TransactionRequest. This can be launched with `await signer.sendTransaction({to: adjudicator.address, ...txRequest}`)
 */
export declare function createConcludeTransaction(conclusionProof: SignedState[]): providers.TransactionRequest;
/**
 * Marshalls the supplied destination, expectedHeld and amount into an ethereum transaction for the deposit ETH method.
 * @param destination The channelId to deposit into
 * @param expectedHeld The amount you expect to have already been deposited
 * @param amount The amount you intend to deposit
 * @returns the transaction request
 */
export declare function createETHDepositTransaction(destination: string, expectedHeld: string, amount: string): providers.TransactionRequest;
/**
 * Marshalls the supplied tokenAddress, destination, expectedHeld and amount into an ethereum transaction for the deposit ERC20 token method.
 * @param tokenAddress The ERC20 token contract address
 * @param destination The channelId to deposit into
 * @param expectedHeld The amount you expect to have already been deposited
 * @param amount The amount you intend to deposit
 * @returns the transaction request
 */
export declare function createERC20DepositTransaction(tokenAddress: string, destination: string, expectedHeld: string, amount: string): providers.TransactionRequest;
/**
 * Marshalls the supplied signedStates into an ethereum transaction for the conclude and transfer all assets method. Automatically computes signedVariableParts, etc.
 * @param signedStates an array of signed states
 * @returns An ethers TransactionRequest. This can be launched with `await signer.sendTransaction({to: adjudicator.address, ...txRequest}`)
 */
export declare function createConcludeAndTransferAllAssetsTransaction(signedStates: SignedState[]): providers.TransactionRequest;
/**
 * Marshalls the supplied signedStates into an ethereum transaction for the transfer all assets method. Automatically computes signedVariableParts, etc.
 * @param state last state in the channel
 * @returns An ethers TransactionRequest. This can be launched with `await signer.sendTransaction({to: adjudicator.address, ...txRequest}`)
 */
export declare function createTransferAllAssetsTransaction(state: State): providers.TransactionRequest;
/**
 * Marshalls the supplied signedStates into the signature arguments used in most on chain methods.
 * Currently we assume each signedState is a unique combination of state/signature
 * So if multiple participants sign a state we expect a SignedState for each participant
 * @param signedStates an array of signed states
 * @returns Object with (states, signatures, whosignedWhat)
 */
export declare function createSignatureArguments(signedStates: SignedState[]): {
    states: State[];
    signatures: Signature[];
    whoSignedWhat: number[];
};
